import math

from acestream.object import Extendable
from acestream.stream import Stream


class ChannelResult(Extendable):

  def __init__(self, server, data):
    Extendable.__init__(self)

    self.total = 0
    self.name  = None
    self.icon  = None
    self.epg   = None
    self.items = None

    self._generate_items(server, data.pop('items'))
    self._set_attrs_to_values(data)
    self._parse_attributes()

  def _generate_items(self, server, results):
    self.items = [StreamResult(server, i) for i in results]

  def _parse_attributes(self):
    self.name  = self.name.strip()
    self.total = len(self.items)


class StreamResult(Extendable):

  def __init__(self, server, data):
    Extendable.__init__(self)

    self.infohash                = None
    self.name                    = None
    self.channel_id              = None
    self.bitrate                 = None
    self.categories              = None
    self.availability_updated_at = None
    self.availability            = None
    self.status                  = None
    self.in_playlist             = None
    self.stream                  = None

    self._generate_stream(server, data.get('infohash'))
    self._set_attrs_to_values(data)
    self._parse_attributes()

  def _generate_stream(self, server, infohash):
    self.stream = Stream(server, infohash=infohash)

  def _parse_attributes(self):
    self.name = self.name.strip()


class Search(Extendable):

  def __init__(self, server, **params):
    Extendable.__init__(self)

    self.total_pages = 0
    self.total       = 0
    self.time        = 0
    self.results     = None
    self.server      = server
    self.params      = params
    self.page        = int(params.pop('page', 1))
    self.page_size   = int(params.pop('page_size', 10))
    self.groups      = bool(params.pop('group_by_channels', False))

  def get(self, page=1):
    self.page = page
    response  = self.server.getsearch(**self.query_params)

    if response.success:
      results = response.data.pop('results')
      self._generate_results(results)

      self._set_attrs_to_values(response.data)
      self.total_pages = math.ceil(self.total / self.page_size)

  @property
  def query_params(self):
    params = self.params

    params['page']              = self.page
    params['page_size']         = self.page_size
    params['group_by_channels'] = self.groups

    return params

  def _generate_results(self, results):
    if self.groups:
      self.results = [ChannelResult(self.server, i) for i in results]
    else:
      self.results = [StreamResult(self.server, i) for i in results]



class Extendable(object):

  def _set_attrs_to_values(self, data={}):
    if isinstance(data, dict):
      for key in data.keys():
        if not hasattr(self, key) or not callable(getattr(self, key)):
          setattr(self, key, data[key])


class Observable(object):

  def __init__(self):
    self._events = []

  def connect(self, event_name, callback_fn):
    self._events.append({ 'event_name': event_name, 'callback_fn': callback_fn })

  def emit(self, event_name, *callback_args):
    for event in self._events:
      if event['event_name'] == event_name:
        event['callback_fn'](*callback_args)



import json

from distutils.version import LooseVersion

try:
  from urllib.request import urlopen
  from urllib.parse import urlencode
  from urllib.error import HTTPError
except ImportError:
  from urllib import urlopen
  from urllib import urlencode
  from urllib2 import HTTPError


class Response(object):

  def __init__(self, data=None, message=None, error=False):
    self.data    = data
    self.error   = error
    self.success = not bool(error)
    self.message = self._parse_message(message)

  def _parse_message(self, message):
    if message:
      message = message.split(']')[-1]
      message = message.lstrip('<').rstrip('>').strip()
      message = '%s%s' % (message[0].upper(), message[1:])

      return message


class Request(object):

  def __init__(self, host, port=None, scheme='http'):
    self.base = self._geturl_base(scheme, host, str(port))

  def get(self, req_url, **params):
    apiurl = self._geturl(req_url, **params)
    return self._request(apiurl)

  def _geturl(self, path, **params):
    params = dict(map(self._parse_param, params.items()))
    params = urlencode(params)
    apiurl = str(path).replace(self.base, '').strip('/')

    return '{0}/{1}?{2}'.format(self.base, apiurl, params)

  def _request(self, req_url):
    try:
      response = self._parse_json(urlopen(req_url).read())
    except (IOError, HTTPError) as error:
      response = { 'result': None, 'error': str(error) }

    return self._generate_response(response)

  def _generate_response(self, output):
    result = output.get('result') or output.get('response')
    error  = output.get('error')

    return Response(data=result, error=bool(error), message=error)

  def _geturl_base(self, scheme, host, port):
    if scheme and not host.startswith(scheme):
      host = '{0}://{1}'.format(scheme, host)

    if port and not host.endswith(port):
      host = '{0}:{1}'.format(host, port)

    return host

  def _get_response_key(self, response, key):
    if response.success:
      return response.data.get(key)

  def _parse_json(self, string):
    try:
      return json.loads(string)
    except (IOError, ValueError):
      return {}

  def _parse_param(self, param):
    key, value = param

    if isinstance(value, bool):
      value = int(value)

    return (key, value)


class Server(Request):

  def __init__(self, host, port=6878, scheme='http', api_token=None):
    Request.__init__(self, host, port, scheme)

    self.api_token = api_token

  def getservice(self, **params):
    return self.get('webui/api/service', format='json', **params)

  def getversion(self):
    return self.getservice(method='get_version')

  def getserver(self, **params):
    return self.get('server/api', **params)

  def gettoken(self):
    return self.getserver(method='get_api_access_token')

  def getsearch(self, **params):
    return self.getserver(method='search', token=self.token, **params)

  def getstream(self, **params):
    is_hls = params.pop('hls', False)
    apiurl = 'manifest.m3u8' if is_hls else 'getstream'

    if LooseVersion(self.version) < LooseVersion('3.1.29'):
      params['sid'] = params.pop('pid')

    return self.get('ace/{0}'.format(apiurl), format='json', **params)

  def getbroadcast(self, manifest_url):
    return self.get('hls/manifest.m3u8', format='json', manifest_url=manifest_url)

  @property
  def version(self):
    response = self.getversion()
    return self._get_response_key(response, 'version')

  @property
  def available(self):
    return bool(self.version)

  @property
  def token(self):
    if not self.api_token:
      response = self.gettoken()
      self.api_token = self._get_response_key(response, 'token')

    return self.api_token



import os
import signal
import subprocess

from threading import Thread
from acestream.object import Observable


class Engine(Observable):

  def __init__(self, bin, **options):
    Observable.__init__(self)

    self.process = None
    self.bin     = bin
    self.options = options

  def start(self, daemon=True, **kwargs):
    if not self.running:
      thread = Thread(target=self._start_process, kwargs=kwargs)
      thread.setDaemon(daemon)
      thread.start()

  def stop(self):
    if self.process:
      os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)

      self.process = None
      self.emit('terminated')

  @property
  def running(self):
    return bool(self.process)

  @property
  def process_args(self):
    options = self.bin.split()

    for (key, value) in self.options.items():
      options.append('--{0}'.format(key.replace('_', '-')))

      if not isinstance(value, bool):
        options.append(str(value))

    return options

  def _start_process(self, **kwargs):
    kwargs['preexec_fn'] = os.setsid

    try:
      self.process = subprocess.Popen(self.process_args, **kwargs)
      self.emit('started')
      self.process.communicate()

      self.process = None
      self.emit('terminated')
    except OSError as error:
      self.process = None
      self.emit('error', str(error))



import setuptools

with open('README.md', 'r') as fh:
  long_description = fh.read()

setuptools.setup(
  name='acestream',
  version='0.1.8',
  author='Jonian Guveli',
  author_email='jonian@hardpixel.eu',
  description='Interact with the AceStream Engine and the HTTP API',
  long_description=long_description,
  long_description_content_type='text/markdown',
  url='https://github.com/jonian/python-acestream',
  packages=setuptools.find_packages(),
  classifiers=[
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 3',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Operating System :: POSIX :: Linux'
  ],
  project_urls={
    'Bug Reports': 'https://github.com/jonian/python-acestream/issues',
    'Source': 'https://github.com/jonian/python-acestream',
  }
)



import time
import hashlib

from threading import Thread
from acestream.object import Extendable
from acestream.object import Observable


class Stats(Extendable, Observable):

  def __init__(self, server):
    Extendable.__init__(self)
    Observable.__init__(self)

    self.stat_url       = None
    self.status         = None
    self.peers          = 0
    self.speed_down     = 0
    self.speed_up       = 0
    self.downloaded     = 0
    self.uploaded       = 0
    self.progress       = 0
    self.total_progress = 0
    self.server         = server

  def watch(self, stat_url):
    self.stat_url = stat_url
    poller_thread = Thread(target=self._poll_stats)

    poller_thread.setDaemon(True)
    poller_thread.start()

  def stop(self):
    self.stat_url = None

  def update(self):
    response = self.server.get(self.stat_url)
    self._set_response_to_values(response)

  def _set_response_to_values(self, response):
    if response.success:
      self._set_attrs_to_values(response.data)
      self.emit('updated')

  def _poll_stats(self):
    while self.stat_url:
      time.sleep(1)
      self.update()


class Stream(Extendable, Observable):

  def __init__(self, server, id=None, url=None, infohash=None):
    Extendable.__init__(self)
    Observable.__init__(self)

    self.status              = None
    self.is_live             = None
    self.playback_session_id = None
    self.command_url         = None
    self.playback_url        = None
    self.stat_url            = None
    self.server              = server
    self.stats               = Stats(server)

    self._check_required_args(id=id, url=url, infohash=infohash)
    self._parse_stream_params(id=id, url=url, infohash=infohash)

  def start(self, hls=False, **kwargs):
    kwparams = dict(kwargs, **self.params) if hls else self.params
    response = self.server.getstream(pid=self.pid, hls=hls, **kwparams)

    if response.success:
      self._set_attrs_to_values(response.data)
      self._start_watchers()

      self.emit('started')
    else:
      self.emit('error', response.message)

  def stop(self):
    response = self.server.get(self.command_url, method='stop')

    if response.success:
      self._stop_watchers()
      self.emit('stopped')
    else:
      self.emit('error', response.message)

  @property
  def params(self):
    params = { 'id': self.id, 'url': self.url, 'infohash': self.infohash }
    params = dict(filter(lambda item: item[1] is not None, params.items()))

    return params

  def _start_watchers(self):
    if self.stat_url:
      self.stats.watch(self.stat_url)
      self.stats.connect('updated', self._on_stats_update)

  def _stop_watchers(self):
    self.stats.stop()

  def _check_required_args(self, **kwargs):
    values = list(filter(None, kwargs.values()))
    params = "'id' or 'url' or 'infohash'"

    if not any(values):
      banner = '__init__() missing 1 required positional argument'
      raise TypeError('{0}: {1}'.format(banner, params))

    if len(values) > 1:
      banner = '__init__() too many positional arguments, provide only one of'
      raise TypeError('{0}: {1}'.format(banner, params))

  def _parse_stream_params(self, **kwargs):
    sid_args = list(filter(None, kwargs.values()))
    self.pid = hashlib.sha1(sid_args[0].encode('utf-8')).hexdigest()

    self._set_attrs_to_values(kwargs)

  def _on_stats_update(self):
    prev_status = self.status
    self.status = self.stats.status

    self.emit('stats::updated')

    if prev_status != self.status:
      self.emit('status::changed')
